// ****************************************************************************
//
//  	CDarla24DspCommObject.cpp
//
//		Implementation file for Darla24 DSP interface class.
//
//		Copyright Echo Digital Audio Corporation (c) 1998 - 2002
//		All rights reserved
//		www.echoaudio.com
//		
//		Permission is hereby granted, free of charge, to any person obtaining a
//		copy of this software and associated documentation files (the
//		"Software"), to deal with the Software without restriction, including
//		without limitation the rights to use, copy, modify, merge, publish,
//		distribute, sublicense, and/or sell copies of the Software, and to
//		permit persons to whom the Software is furnished to do so, subject to
//		the following conditions:
//		
//		- Redistributions of source code must retain the above copyright
//		notice, this list of conditions and the following disclaimers.
//		
//		- Redistributions in binary form must reproduce the above copyright
//		notice, this list of conditions and the following disclaimers in the
//		documentation and/or other materials provided with the distribution.
//		
//		- Neither the name of Echo Digital Audio, nor the names of its
//		contributors may be used to endorse or promote products derived from
//		this Software without specific prior written permission.
//
//		THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//		EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//		MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//		IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
//		ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//		TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//		SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************

#define ECHOGALS_FAMILY
#include "Echoaudio.h"
#include "Echo_Dsp.h"

#include "DSP/Darla24DSP.c"


/****************************************************************************

	Magic constants for the Darla24 hardware

 ****************************************************************************/

#define GD24_96000      0x0
#define GD24_48000      0x1
#define GD24_44100      0x2
#define GD24_32000      0x3
#define GD24_22050      0x4
#define GD24_16000      0x5
#define GD24_11025      0x6
#define GD24_8000       0x7
#define GD24_88200      0x8
#define GD24_EXT_SYNC   0x9




//===========================================================================
//
// Constructor
//
//===========================================================================

int NewDspComm(struct CEchoGals *eg) {

  ECHO_DEBUGPRINTF(("NewDspComm()\n"));

  // Check if the compiler added extra padding inside the structure
  if ((long)&((DspCommPage *)0)->byVmixerLevel != 0xba0L) {
    ECHO_DEBUGPRINTF(("NewDspComm() - Invalid DspComm structure\n"));
    return(-1);
  }
  if (sizeof(dma_addr_t)!=4) {
    ECHO_DEBUGPRINTF(("NewDspComm() - dma_addr_t must be 32 bits long\n"));
    return(-1);
  }

  // This part is common to all the cards
  CreateCommPage(eg, eg->m_pvSharedMemory);     // Virtual addr DSP's register base

  // Write the features of the card
  eg->m_wNumPipesOut = 8;
  eg->m_wNumPipesIn = 2;
  eg->m_wNumBussesOut = 8;
  eg->m_wNumBussesIn = 2;
  eg->m_wFirstDigitalBusOut = 8;
  eg->m_wFirstDigitalBusIn = 2;
  eg->m_fHasVmixer = FALSE;
  eg->m_wNumMidiOut = 0;                        // # MIDI out channels
  eg->m_wNumMidiIn = 0;                         // # MIDI in  channels
  eg->m_pDspCommPage->dwSampleRate = SWAP32((u32)44100);        // Need this in case we start with ESYNC
  eg->m_pwDspCodeToLoad=pwDarla24DSP;
  // Since this card has no ASIC, mark it as loaded so everything works OK
  eg->m_bASICLoaded = TRUE;
  ECHO_DEBUGPRINTF(("NewDspComm() done\n"));
  return(0);
}



// The Gina20 has no ASIC. Just do nothing
int LoadASIC(struct CEchoGals *eg) {

  return(TRUE);
}



//===========================================================================
//
// SetSampleRate
// 
// Set the audio sample rate for Darla24; this is fairly simple.  You
// just pick the right magic number.
//
//===========================================================================

u32 SetSampleRate(struct CEchoGals *eg, u32 SampleRate) {
  u8 Clock;

  // Pick the magic number
  switch (SampleRate) {
    case 96000:
      Clock = GD24_96000;
      break;
    case 88200:
      Clock = GD24_88200;
      break;
    case 48000:
      Clock = GD24_48000;
      break;
    case 44100:
      Clock = GD24_44100;
      break;
    case 32000:
      Clock = GD24_32000;
      break;
    case 22050:
      Clock = GD24_22050;
      break;
    case 16000:
      Clock = GD24_16000;
      break;
    case 11025:
      Clock = GD24_11025;
      break;
    case 8000:
      Clock = GD24_8000;
      break;
    default:
      ECHO_DEBUGPRINTF(("SetSampleRate: Error, invalid sample rate 0x%lx\n", SampleRate));
      return(0xffffffff);
  }

  if (!WaitForHandshake(eg))
    return(0xffffffff);

  // Override the sample rate if this card is set to Echo sync.
  // m_pDspCommPage->wInputClock is just being used as a parameter here;
  // the DSP ignores it.
  if (eg->InputClock==ECHO_CLOCK_ESYNC)
    Clock = GD24_EXT_SYNC;

  eg->m_pDspCommPage->dwSampleRate = SWAP32(SampleRate);

  // Write the audio state to the comm page
  eg->m_pDspCommPage->byGDClockState = Clock;

  // Send command to DSP
  ClearHandshake(eg);
  SendVector(eg, DSP_VC_SET_GD_AUDIO_STATE);

  ECHO_DEBUGPRINTF(("SetSampleRate: 0x%lx clocks %s\n", SampleRate));
  return(GetSampleRate(eg));
}


//===========================================================================
//
// Set input clock
//
// Darla24 supports internal and Esync clock.
//
//===========================================================================

int SetInputClock(struct CEchoGals *eg, u16 Clock) {
  if (Clock!=ECHO_CLOCK_INTERNAL && Clock!=ECHO_CLOCK_ESYNC)
    return(ECHOSTATUS_CLOCK_NOT_SUPPORTED);

  eg->InputClock = Clock;

  return(SetSampleRate(eg, GetSampleRate(eg)));
}



// Layla20 only
int SetOutputClock(struct CEchoGals *eg, u16 Clock) {

  return(ECHOSTATUS_CLOCK_NOT_SUPPORTED);
}



// This card has no vmixer.
int SetVmixerGain(struct CEchoGals *eg, u16 Pipe, u16 Output, int Gain) {

  return(ECHOSTATUS_NOT_SUPPORTED);
}



// Gina24, Mona, Layla24 - only
int DspSetDigitalMode(struct CEchoGals *eg, u8 Mode) {

  return(ECHOSTATUS_DIGITAL_MODE_NOT_SUPPORTED);
}



// This card has no vmixer.
int UpdateVmixerLevel(struct CEchoGals *eg) {

  return(ECHOSTATUS_NOT_SUPPORTED);
}



// This card has no digital i/o
int SetProfessionalSpdif(struct CEchoGals *eg, char Prof) {

  return(ECHOSTATUS_NOT_SUPPORTED);
}
