// ****************************************************************************
//
//      CMiaDspCommObject.cpp
//
//              Implementation file for EchoGals generic driver Darla 24 DSP
//              interface class.
//
//              Copyright Echo Digital Audio Corporation (c) 1998 - 2002
//              All rights reserved
//              www.echoaudio.com
//              
//              Permission is hereby granted, free of charge, to any person obtaining a
//              copy of this software and associated documentation files (the
//              "Software"), to deal with the Software without restriction, including
//              without limitation the rights to use, copy, modify, merge, publish,
//              distribute, sublicense, and/or sell copies of the Software, and to
//              permit persons to whom the Software is furnished to do so, subject to
//              the following conditions:
//              
//              - Redistributions of source code must retain the above copyright
//              notice, this list of conditions and the following disclaimers.
//              
//              - Redistributions in binary form must reproduce the above copyright
//              notice, this list of conditions and the following disclaimers in the
//              documentation and/or other materials provided with the distribution.
//              
//              - Neither the name of Echo Digital Audio, nor the names of its
//              contributors may be used to endorse or promote products derived from
//              this Software without specific prior written permission.
//
//              THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//              EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//              MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//              IN NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR
//              ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//              TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//              SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
//
// ****************************************************************************

#define ECHO24_FAMILY
#include "Echoaudio.h"
#include "Echo_Dsp.h"

#include "DSP/MiaDSP.c"


/****************************************************************************

	Construction and destruction

 ****************************************************************************/

//===========================================================================
//
// Constructor
//
//===========================================================================

int NewDspComm(struct CEchoGals *eg) {

  ECHO_DEBUGPRINTF(("NewDspComm()\n"));

  // Check if the compiler added extra padding inside the structure
  if ((long)&((DspCommPage *)0)->byVmixerLevel != 0xba0L) {
    ECHO_DEBUGPRINTF(("NewDspComm() - Invalid DspComm structure\n"));
    return(-1);
  }
  if (sizeof(dma_addr_t)!=4) {
    ECHO_DEBUGPRINTF(("NewDspComm() - dma_addr_t must be 32 bits long\n"));
    return(-1);
  }

  // This part is common to all the cards
  CreateCommPage(eg, eg->m_pvSharedMemory);     // Virtual addr DSP's register base

  // Write the features of the card
  eg->m_wNumPipesOut = 8;
  eg->m_wNumPipesIn = 4;
  eg->m_wNumBussesOut = 4;
  eg->m_wNumBussesIn = 4;
  eg->m_wFirstDigitalBusOut = 2;
  eg->m_wFirstDigitalBusIn = 2;
  eg->m_fHasVmixer = TRUE;
  eg->m_wNumMidiOut = 0;                        // # MIDI out channels
  eg->m_wNumMidiIn = 0;                         // # MIDI in  channels
  eg->m_pDspCommPage->dwSampleRate = SWAP32((u32)44100);
  eg->m_pwDspCodeToLoad = pwMiaDSP;
  // Since this card has no ASIC, mark it as loaded so everything works OK
  eg->m_bASICLoaded = TRUE;
  ECHO_DEBUGPRINTF(("NewDspComm() done\n"));
  return(0);
}



// The Mia has no ASIC. Just do nothing
int LoadASIC(struct CEchoGals *eg) {

  return(TRUE);
}



/****************************************************************************

	Hardware setup and config

 ****************************************************************************/

//===========================================================================
//
//      Set the input clock
//
//===========================================================================

int SetInputClock(struct CEchoGals *eg, u16 Clock) {
  u32 SampleRate;

  ECHO_DEBUGPRINTF(("SetInputClock\n"));
  SampleRate = GetSampleRate(eg);

  switch (Clock) {
    case ECHO_CLOCK_INTERNAL:
      ECHO_DEBUGPRINTF(("Set Mia clock to INTERNAL\n"));
      // If the sample rate is out of range for some reason, set it to a reasonable value.  mattg
      if (SampleRate<8000 || SampleRate>96000)
        SampleRate = 48000;
      break;
    case ECHO_CLOCK_SPDIF:
      ECHO_DEBUGPRINTF(("Set Mia clock to SPDIF\n"));
      break;
    default:
      ECHO_DEBUGPRINTF(("Input clock 0x%x not supported for Mia\n", Clock));
      return(ECHOSTATUS_CLOCK_NOT_SUPPORTED);
  }

  eg->InputClock = Clock;
  SetSampleRate(eg, SampleRate);
  return(ECHOSTATUS_OK);
}


//===========================================================================
//
// SetSampleRate
//
// Set the audio sample rate for Mia
//
//===========================================================================


u32 SetSampleRate(struct CEchoGals *eg, u32 SampleRate) {
  u32 ControlReg;

  ControlReg = MIA_48000;

  switch (SampleRate) {
    case 96000:
      ControlReg = MIA_96000;
      break;
    case 88200:
      ControlReg = MIA_88200;
      break;
    case 44100:
      ControlReg = MIA_44100;
      break;
    case 32000:
      ControlReg = MIA_32000;
      break;
    case 24000:
      ControlReg = MIA_24000;
      break;
    case 22050:
      ControlReg = MIA_22050;
      break;
    case 16000:
      ControlReg = MIA_16000;
      break;
    case 12000:
      ControlReg = MIA_12000;
      break;
    case 11025:
      ControlReg = MIA_11025;
      break;
    case 8000:
      ControlReg = MIA_8000;
      break;
  }

  // Override the clock setting if this Mia is set to S/PDIF clock
  if (eg->InputClock==ECHO_CLOCK_SPDIF) {
    ControlReg &= MIA_SRC_MASK;
    ControlReg |= MIA_SPDIF;
  }

  // Set the control register if it has changed
  if (ControlReg!=GetControlRegister(eg)) {
    if (!WaitForHandshake(eg))
      return(0xffffffff);

    // Set the values in the comm page; the dwSampleRate field isn't (@@ interesting)
    // used by the DSP, but is read by the call to GetSampleRate below
    eg->m_pDspCommPage->dwSampleRate = SWAP32(SampleRate);
    SetControlRegister(eg, ControlReg);

    // Poke the DSP
    ClearHandshake(eg);
    SendVector(eg, DSP_VC_UPDATE_CLOCKS);
  }

  return(GetSampleRate(eg));
}


/*
//===========================================================================
//
// GetPipeOutGain and SetPipeOutGain
//
// On Mia, this doesn't set the line out volume; instead, it sets the vmixer volume.
//
//===========================================================================

ECHOSTATUS CMiaDspCommObject::SetPipeOutGain(WORD wPipeOut, WORD wBusOut, int iGain, BOOL fImmediate)
{
  if (wPipeOut >= m_wNumPipesOut) {
    ECHO_DEBUGPRINTF(("CDspCommObject::SetPipeOutGain: Invalid out pipe " "%d\n", wPipeOut));

    return ECHOSTATUS_INVALID_CHANNEL;
  }

  iGain = GENERIC_TO_DSP(iGain);

  if (wBusOut < m_wNumBussesOut) {
    if (!WaitForHandshake())
      return ECHOSTATUS_DSP_DEAD;

    DWORD dwIndex = wBusOut * m_wNumPipesOut + wPipeOut;

    m_pDspCommPage->byVmixerLevel[dwIndex] = (BYTE) iGain;

    ECHO_DEBUGPRINTF(("CMiaDspCommObject::SetPipeOutGain: Out pipe %d, " "out bus %d = %u\n", wPipeOut, wBusOut, iGain));

    if (fImmediate) {
      return UpdateVmixerLevel();
    }

    return ECHOSTATUS_OK;
  }

  ECHO_DEBUGPRINTF(("CMiaDspCommObject::SetPipeOutGain: Invalid out bus " "%d\n", wBusOut));

  return ECHOSTATUS_INVALID_CHANNEL;

}                               // SetPipeOutGain


ECHOSTATUS CMiaDspCommObject::GetPipeOutGain(WORD wPipeOut, WORD wBusOut, int &iGain)
{
  if (wPipeOut >= m_wNumPipesOut) {
    ECHO_DEBUGPRINTF(("CMiaDspCommObject::GetPipeOutGain: Invalid out pipe " "%d\n", wPipeOut));

    return ECHOSTATUS_INVALID_CHANNEL;
  }

  if (wBusOut < m_wNumBussesOut) {
    iGain = m_pDspCommPage->byVmixerLevel[wBusOut * m_wNumPipesOut + wPipeOut];
    iGain = DSP_TO_GENERIC(iGain);
    return ECHOSTATUS_OK;
  }

  ECHO_DEBUGPRINTF(("CMiaDspCommObject::GetPipeOutGain: Invalid out bus " "%d\n", wBusOut));

  return ECHOSTATUS_INVALID_CHANNEL;

}                               // GetPipeOutGain
*/




// Layla20 only
int SetOutputClock(struct CEchoGals *eg, u16 Clock) {
  
  return(ECHOSTATUS_CLOCK_NOT_SUPPORTED);
}



// Gina24, Mona, Layla24 - only
int DspSetDigitalMode(struct CEchoGals *eg, u8 Mode) {

  return(ECHOSTATUS_DIGITAL_MODE_NOT_SUPPORTED);
}



// This function routes the sound from a virtual channel to a real output
int SetVmixerGain(struct CEchoGals *eg, u16 Pipe, u16 Output, int Gain) {
  int Index;

  if (Pipe >= eg->m_wNumPipesOut) {
    ECHO_DEBUGPRINTF(("SetVmixertGain: Invalid Pipe %d\n", Pipe));
    return(ECHOSTATUS_INVALID_CHANNEL);
  }

  if (Output >= eg->m_wNumBussesOut) {
    ECHO_DEBUGPRINTF(("SetVmixerGain: Invalid out %d\n", (int)Output));
    return(ECHOSTATUS_INVALID_CHANNEL);
  }

  if (!WaitForHandshake(eg))
    return(ECHOSTATUS_DSP_DEAD);

  eg->VmixerLevel[Pipe][Output] = Gain;
  Index = Output * eg->m_wNumPipesOut + Pipe;
  Gain = GENERIC_TO_DSP(Gain);
  eg->m_pDspCommPage->byVmixerLevel[Index] = Gain;

  ECHO_DEBUGPRINTF(("SetVmixerGain: pipe %d, out %d = %d\n", (int)Pipe, (int)Output, Gain));
  return(ECHOSTATUS_OK);
}



// Tell the DSP to read and update virtual mixer levels in comm page.
int UpdateVmixerLevel(struct CEchoGals *eg) {

  ECHO_DEBUGPRINTF(("UpdateVmixerLevel\n"));

  ClearHandshake(eg);
  return(SendVector(eg, DSP_VC_SET_VMIXER_GAIN));
}



int SetProfessionalSpdif(struct CEchoGals *eg, char Prof) {

  ECHO_DEBUGPRINTF(("SetProfessionalSpdif %d\n", Prof));
  if (Prof)
    DspSetFlags(eg, DSP_FLAG_PROFESSIONAL_SPDIF);
  else
    DspClearFlags(eg, DSP_FLAG_PROFESSIONAL_SPDIF);
  return(ECHOSTATUS_OK);
}


